@@ -269,7 +269,11 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
         }
 
         let stage = device.prepare_stage(data_size)?;
-        unsafe { stage.write(&device.raw, 0, data) }.map_err(DeviceError::from)?;
+        unsafe {
+            profiling::scope!("copy");
+            stage.write(&device.raw, 0, data)
+        }
+        .map_err(DeviceError::from)?;
 
         let mut trackers = device.trackers.lock();
         let (dst, transition) = trackers
@@ -445,9 +449,8 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
         let mapping = unsafe { device.raw.map_buffer(&stage.buffer, 0..stage_size) }
             .map_err(DeviceError::from)?;
         unsafe {
-            profiling::scope!("copy");
-
             if stage_bytes_per_row == bytes_per_row {
+                profiling::scope!("copy aligned");
                 // Fast path if the data is already being aligned optimally.
                 ptr::copy_nonoverlapping(
                     data.as_ptr().offset(data_layout.offset as isize),
@@ -455,6 +458,7 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
                     stage_size as usize,
                 );
             } else {
+                profiling::scope!("copy chunked");
                 // Copy row by row into the optimal alignment.
                 let copy_bytes_per_row = stage_bytes_per_row.min(bytes_per_row) as usize;
                 for layer in 0..size.depth_or_array_layers {
@@ -555,7 +559,8 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
                     let (compute_pipe_guard, mut token) = hub.compute_pipelines.read(&mut token);
                     let (render_pipe_guard, mut token) = hub.render_pipelines.read(&mut token);
                     let (mut buffer_guard, mut token) = hub.buffers.write(&mut token);
-                    let (texture_guard, mut token) = hub.textures.write(&mut token);
+                    // This could be made immutable. It's only mutated for the `has_work` flag.
+                    let (mut texture_guard, mut token) = hub.textures.write(&mut token);
                     let (texture_view_guard, mut token) = hub.texture_views.read(&mut token);
                     let (sampler_guard, mut token) = hub.samplers.read(&mut token);
                     let (query_set_guard, _) = hub.query_sets.read(&mut token);
@@ -615,14 +620,18 @@ impl<G: GlobalIdentityHandlerFactory> Global<G> {
                             }
                         }
                         for id in cmdbuf.trackers.textures.used() {
-                            let texture = &texture_guard[id];
+                            let texture = &mut texture_guard[id];
                             match texture.inner {
                                 TextureInner::Native { raw: None } => {
                                     return Err(QueueSubmitError::DestroyedTexture(id.0));
                                 }
                                 TextureInner::Native { raw: Some(_) } => {}
-                                TextureInner::Surface { .. } => {
+                                TextureInner::Surface {
+                                    ref mut has_work, ..
+                                } => {
                                     use track::ResourceState as _;
+
+                                    *has_work = true;
                                     let ref_count = cmdbuf.trackers.textures.get_ref_count(id);
                                     //TODO: better error handling here?
                                     {
