@@ -37,17 +37,17 @@ pub mod device;
 pub mod error;
 pub mod hub;
 pub mod id;
+mod init_tracker;
 pub mod instance;
-mod memory_init_tracker;
 pub mod pipeline;
 pub mod present;
 pub mod resource;
 mod track;
 mod validation;
 
-pub use hal::api;
+pub use hal::{api, MAX_BIND_GROUPS, MAX_COLOR_TARGETS, MAX_VERTEX_BUFFERS};
 
-use atomic::{AtomicU64, AtomicUsize, Ordering};
+use atomic::{AtomicUsize, Ordering};
 
 use std::{borrow::Cow, os::raw::c_char, ptr, sync::atomic};
 
@@ -147,7 +147,7 @@ impl Drop for MultiRefCount {
 #[derive(Debug)]
 pub struct LifeGuard {
     ref_count: Option<RefCount>,
-    submission_index: AtomicU64,
+    submission_index: AtomicUsize,
     #[cfg(debug_assertions)]
     pub(crate) label: String,
 }
@@ -158,7 +158,7 @@ impl LifeGuard {
         let bx = Box::new(AtomicUsize::new(1));
         Self {
             ref_count: ptr::NonNull::new(Box::into_raw(bx)).map(RefCount),
-            submission_index: AtomicU64::new(0),
+            submission_index: AtomicUsize::new(0),
             #[cfg(debug_assertions)]
             label: label.to_string(),
         }
@@ -170,9 +170,14 @@ impl LifeGuard {
 
     /// Returns `true` if the resource is still needed by the user.
     fn use_at(&self, submit_index: SubmissionIndex) -> bool {
-        self.submission_index.store(submit_index, Ordering::Release);
+        self.submission_index
+            .store(submit_index as _, Ordering::Release);
         self.ref_count.is_some()
     }
+
+    fn life_count(&self) -> SubmissionIndex {
+        self.submission_index.load(Ordering::Acquire) as _
+    }
 }
 
 #[derive(Clone, Debug)]
@@ -206,7 +211,10 @@ macro_rules! gfx_select {
             wgt::Backend::Dx12 => $global.$method::<$crate::api::Dx12>( $($param),* ),
             //#[cfg(all(not(target_arch = "wasm32"), windows))]
             //wgt::Backend::Dx11 => $global.$method::<$crate::api::Dx11>( $($param),* ),
-            #[cfg(all(not(target_arch = "wasm32"), unix, not(any(target_os = "ios", target_os = "macos"))))]
+            #[cfg(any(
+                all(unix, not(target_os = "macos"), not(target_os = "ios")),
+                target_arch = "wasm32"
+            ))]
             wgt::Backend::Gl => $global.$method::<$crate::api::Gles>( $($param),+ ),
             other => panic!("Unexpected backend {:?}", other),
 
