@@ -1,8 +1,8 @@
 use crate::{
     device::{DeviceError, HostMap, MissingFeatures},
-    hub::Resource,
+    hub::{Global, GlobalIdentityHandlerFactory, HalApi, Resource, Token},
     id::{DeviceId, SurfaceId, TextureId, Valid},
-    memory_init_tracker::MemoryInitTracker,
+    init_tracker::BufferInitTracker,
     track::{TextureSelector, DUMMY_SELECTOR},
     validation::MissingBufferUsageError,
     Label, LifeGuard, RefCount, Stored,
@@ -109,7 +109,7 @@ pub(crate) struct BufferPendingMapping {
     pub range: Range<wgt::BufferAddress>,
     pub op: BufferMapOperation,
     // hold the parent alive while the mapping is active
-    pub parent_ref_count: RefCount,
+    pub _parent_ref_count: RefCount,
 }
 
 pub type BufferDescriptor<'a> = wgt::BufferDescriptor<Label<'a>>;
@@ -120,7 +120,7 @@ pub struct Buffer<A: hal::Api> {
     pub(crate) device_id: Stored<DeviceId>,
     pub(crate) usage: wgt::BufferUsages,
     pub(crate) size: wgt::BufferAddress,
-    pub(crate) initialization_status: MemoryInitTracker,
+    pub(crate) initialization_status: BufferInitTracker,
     pub(crate) sync_mapped_writes: Option<hal::MemoryRange>,
     pub(crate) life_guard: LifeGuard,
     pub(crate) map_state: BufferMapState<A>,
@@ -164,18 +164,16 @@ pub(crate) enum TextureInner<A: hal::Api> {
     Surface {
         raw: A::SurfaceTexture,
         parent_id: Valid<SurfaceId>,
+        has_work: bool,
     },
 }
 
 impl<A: hal::Api> TextureInner<A> {
     pub fn as_raw(&self) -> Option<&A::Texture> {
         match *self {
-            TextureInner::Native { raw: Some(ref tex) } => Some(tex),
-            TextureInner::Native { raw: None } => None,
-            TextureInner::Surface {
-                ref raw,
-                parent_id: _,
-            } => Some(std::borrow::Borrow::borrow(raw)),
+            Self::Native { raw: Some(ref tex) } => Some(tex),
+            Self::Native { raw: None } => None,
+            Self::Surface { ref raw, .. } => Some(raw.borrow()),
         }
     }
 }
@@ -191,6 +189,27 @@ pub struct Texture<A: hal::Api> {
     pub(crate) life_guard: LifeGuard,
 }
 
+impl<G: GlobalIdentityHandlerFactory> Global<G> {
+    /// # Safety
+    ///
+    /// - The raw texture handle must not be manually destroyed
+    pub unsafe fn texture_as_hal<A: HalApi, F: FnOnce(Option<&A::Texture>)>(
+        &self,
+        id: TextureId,
+        hal_texture_callback: F,
+    ) {
+        profiling::scope!("as_hal", "Texture");
+
+        let hub = A::hub(self);
+        let mut token = Token::root();
+        let (guard, _) = hub.textures.read(&mut token);
+        let texture = guard.get(id).ok();
+        let hal_texture = texture.map(|tex| tex.inner.as_raw().unwrap());
+
+        hal_texture_callback(hal_texture);
+    }
+}
+
 #[derive(Clone, Copy, Debug)]
 pub enum TextureErrorDimension {
     X,
@@ -277,7 +296,10 @@ impl HalTextureViewDescriptor {
 #[derive(Debug)]
 pub struct TextureView<A: hal::Api> {
     pub(crate) raw: A::TextureView,
+    // The parent's refcount is held alive, but the parent may still be deleted
+    // if it's a surface texture. TODO: make this cleaner.
     pub(crate) parent_id: Stored<TextureId>,
+    pub(crate) device_id: Stored<DeviceId>,
     //TODO: store device_id for quick access?
     pub(crate) desc: HalTextureViewDescriptor,
     pub(crate) format_features: wgt::TextureFormatFeatures,
